/******************** (C) COPYRIGHT 2020 Geek************************************
* File Name          : pdr_location.c
* Current Version    : V1.2
* Author             : logzhan
* Date of Issued     : 2020.7.3
* Comments           : PDR㷨
********************************************************************************/
/* Header File Including -----------------------------------------------------*/
#include <stdio.h>
#include <math.h> 
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <stdlib.h>
#include <stdbool.h>
#include "Location.h"
#include "Kalman.h"
#include "Pedometer.h"
#include "AHRS.h"
#include "Detector.h"
#include "Utils.h"
#include "SceneRecognition.h"
#include "CoordTrans.h"

/* Global Variable Definition ------------------------------------------------*/
PDR_t*          pdr;                                  // PDRȫϢ
StepPredict     stepPredict = { 0 };
Classifer_t     sys;
GNSS_t          pgnss;                                // λϢ
EKFPara_t       EkfPara;
double          GpsPos[HIST_GPS_NUM][3] = {{0}};
static uint32_t g_status;
extern AHRS_t   Ahrs;

double refGpsYaw    = -1000;
double yaw_bias     = -1;

float dir[2] = { 0 };
float bias_dir[2] = { 0 };

/**---------------------------------------------------------------------
* Function    : PDRNav_Init
* Description : PDRϵͳʼ
* Date        : 2022/09/21 logzhan
*---------------------------------------------------------------------**/
void PDR_Init(void) {

	memset(&EkfPara,    0, sizeof(EkfPara));
	memset(&pgnss,       0, sizeof(pgnss));
	memset(GpsPos,       0, sizeof(GpsPos));
	memset(&stepPredict, 0, sizeof(stepPredict));
	memset(&sys,         0, sizeof(sys));
	memset(dir,          0, sizeof(dir));
	memset(bias_dir,     0, sizeof(bias_dir));

	// GNSSλʼ
    InitGnssInfo();
    // ʼAHRS̬
    AHRS_Init();
    // ʼ
    Detector_Init();
    // PDRʼ
    pdr = Base_Init();
	//baseļ̬Ǽʼ
	EKF_Init();
    // ʶģʼ
    SceneRecog_Init();

	yaw_bias = -1;
}

/**---------------------------------------------------------------------
* Function    : InitGnssInfo
* Description : PDS GNSSλϢʼ
* Date        : 2020/02/01 logzhan:
*               ʼ汾
*               2020/02/08: logzhan ޸initGpsΪinitGnssInfo
*               ӷϺ
*---------------------------------------------------------------------**/
void InitGnssInfo(void) {
    pgnss.lastError    = ACCURACY_ERR_MAX;
    pgnss.overVelCount = 0;
}

/**---------------------------------------------------------------------
* Function    : ResetSystemStatus
* Description : PDR ϵͳ״̬
* Date        : 2022/09/16 logzhan
*---------------------------------------------------------------------**/
void ResetSystemStatus(EKFPara_t *kf) 
{
    memset(kf, 0, sizeof(EKFPara_t));
    pdr->fusionPdrFlg = PDR_FALSE;
    pdr->sysStatus    = IS_INITIAL;
    pgnss.overVelCount = 0;
}

/**---------------------------------------------------------------------
* Function    : InsLocationUpdate
* Description : PDR Եλ
* Date        : 2022/09/16 logzhan
*---------------------------------------------------------------------**/
int InsLocationUpdate(IMU_t* ImuData, EKFPara_t* Ekf) {

    // AHRS̬ռ2%ʱ
    if (UpdateAHRS(ImuData)) {
        // ̬ƺPCA ռ26%ʱ
        ComputePCA(&Ahrs);
    }
    // û˶ͼIMUϢԼû˶ͼ, ռ21%ʱ
    DetectorUpdateIMU(ImuData);

    pdr->gyroTime = ImuData->gyr.time;
    // Ϊʲôֳ0.1˶Ƶʣ
    if (pdr->MotionType == DETECTOR_TYPE_HANDHELD)pdr->motionFreq = 0.1f;

    // Ʋ봫ݣռ10%ʱ
    PedometerUpdate(ImuData, &pdr->steps);

    // ----- Filter Initialization ------- //
    if (pdr->sysStatus == IS_INITIAL) {
        pdr->ts = ImuData->gyr.time;
        pdr->lastSteps = pdr->steps;
        stepPredict.lastTime = 0;
        stepPredict.fnum = 0;
        stepPredict.fsum = 0;
        return 0;
    }

    pdr->ts = ImuData->gyr.time;

    if ((0 == ImuData->acc.s[0] && 0 == ImuData->acc.s[1] && 0 == ImuData->acc.s[2]) ||
        (0 == ImuData->mag.s[0] && 0 == ImuData->mag.s[1] && 0 == ImuData->mag.s[2])) {
        return 0;
    }

    // ʶǰ״̬Ƿƽȣϵͳ10%ļ, ԺŻ
    // stateRecognition(ss_data->acc.s, &sys);

    pdr->heading = calPredAngle(pdr, &sys, pdr->GpsHeading, pdr->trackHeading, ImuData);
    if (yaw_bias != -1) {
        if (pdr->pca_accuracy > 0 && pdr->pca_accuracy < 0.1 && pdr->MotionType == 3 &&
            (pdr->bias_accuracy > 1.5 || pdr->bias_accuracy == -1 || pgnss.error > 30)
            && (fabs(yaw_bias - 2 * PI) < YAW_THRES * PI / 180 || fabs(yaw_bias - PI) < YAW_THRES * PI / 180 ||
                fabs(yaw_bias) < YAW_THRES * PI / 180)) {
            dir[0] = pdr->pca_direction[0] * bias_dir[0] + pdr->pca_direction[1] * bias_dir[1];
            dir[1] = pdr->pca_direction[1] * bias_dir[0] - pdr->pca_direction[0] * bias_dir[1];
            pdr->heading = atan2(dir[1], dir[0]);
            if (pdr->heading < 0) {
                pdr->heading += TWO_PI;
            }
        }
    }

    double imuYaw = - Ahrs.Yaw;
    if (imuYaw < 0) {
        imuYaw += 360;
    }

    pdr->heading = imuYaw / (180 / 3.14159265);
    if (fabs(pdr->insHeadingOffset) > 0.000001) {
        imuYaw = imuYaw - pdr->insHeadingOffset * (180 / 3.14159265);
        pdr->heading = imuYaw / (180 / 3.14159265);
    }

    // PDR λԤ
    InsLocationPredict(pdr, &stepPredict, ImuData, &pgnss, Ekf);

    pdr->lastSteps   = pdr->steps;
    pdr->lastHeading = pdr->heading;
    return 0;
}

/**----------------------------------------------------------------------
* Function    : GnssInsFusionLocation
* Description : PDR GPSںINSԶλ
* Date        : 2021/01/29 logzhan 
*---------------------------------------------------------------------**/
int GnssInsFusionLocation(Nmea_t *nmea, EKFPara_t *kf, LctFs_t *result)
{
    Nmea2Gnss(nmea, &pgnss);
    // ʱǷ
    if (fabs(pgnss.timestamps - pdr->ts) >= 1000)return TYPE_FIX_NONE;

	// nmeaʶ𳡾Ƿǿճ
	pdr->sceneType = isOpenArea(nmea);
    // HDOPٶȡжgps yawǷ
   
    pdr->gpsSpeed = pgnss.vel * GPS_SPEED_UNIT;
    pdr->lastGpsHeading = pdr->GpsHeading;
    pdr->GpsHeading = GnssCalHeading(&pgnss);
    
    // win32ƽ̨·,
#ifdef _WIN32
    result->pdrHeading = kf->pXk[3];
    result->gpsHeading = pdr->GpsHeading;
    result->hdop       = pgnss.HDOP;
    result->accuracy   = nmea->accuracy.err;
    result->gpsSpeed   = pdr->gpsSpeed;
    result->motionType = pdr->MotionType;
#endif

    // NmeaϢIns
    CalPdrHeadingOffset(nmea, pdr);
    
    // GPSʱ֮䣬仯
    pdr->deltaStepsPerGps = pdr->steps - pdr->lastGpsSteps;
    pdr->lastGpsSteps = pdr->steps;

    int mode = DetectFixMode(&pgnss, kf, pdr, result);
    if (mode != TYPE_FIX_PDR)return mode;

	//ʷGPSԤGPS켣
    pdr->trackHeading = CalGnssTrackHeading(GpsPos, pgnss);
	// ϵͳûгʼȳʼ
    if (pdr->sysStatus == IS_INITIAL) {
        // Gnssϵͳʼ׼
        return GnssInsInitialAlignment(&pgnss, kf, pdr, result);
    }
    // ߵûдﵽԤ켣
    if (!pdr->fusionPdrFlg) 
    {
        return ResetOutputLoction(&pgnss, pdr, kf, result);
    }

    // GPSϢ, kfĹ͹۲
    EKFCalQRMatrix(nmea, pdr, &sys, kf);

    // пλں
    if (pgnss.satellites_num > 4){
        // GPSǴЧ״̬ٶҲЧ״̬
        if (pgnss.yaw == INVAILD_GPS_YAW || pgnss.vel <= 0.0) {
            //if(pdr->heading != INVAILD_GPS_YAW){
            //    pgnss.yaw = (float)(R2D(pdr->heading));
            //}else if (pdr->trackHeading != INVAILD_GPS_YAW) {
            //    pgnss.yaw = (float)(R2D(pdr->trackHeading));
            //}
            pgnss.yaw = (float)kf->pXk[3];
        }
        if (pgnss.yaw != INVAILD_GPS_YAW) {
            EkfGnssInsLocFusion(&pgnss, pdr, &sys, yaw_bias, kf, result);
            return TYPE_FIX_PDR;
        }
    }
    
    // ֱԭʼGPS
    if (pgnss.satellites_num > 4 && nmea->accuracy.err > 0 &&
        nmea->accuracy.err < 15) {
        OutputRawGnssPos(&pgnss, result);
        pdr->NoGnssCount = 0;
        return TYPE_FIX_GPS;
    }

    // ûGPSźţôPDRҲλ, 10
    if (pdr->NoGnssCount < MAX_NO_GPS_PREDICT) {
        // GPS״̬λ
        NoGnssInfoPredict(kf, result, pdr);
        return TYPE_FIX_PDR;
    }
    pdr->NoGnssCount++;
	return 0;
}

/**----------------------------------------------------------------------
* Function    : NoGnssInfoPredict
* Description : ûgpsϢʱԤGPSλ
* Date        : 2022/09/16 
*---------------------------------------------------------------------**/
void NoGnssInfoPredict(EKFPara_t* kf, LctFs_t* result, PDR_t* p_pdr)
{
    double ned[3] = {0}, lla[3] = {0};
    ned[0] = kf->pXk[0] - p_pdr->x0;
    ned[1] = kf->pXk[1] - p_pdr->y0;
    ned[2] = 0;

    NED2WGS84(p_pdr->PllaInit, ned, lla);

    result->latitude        = lla[0];
    result->latitude_ns     = pgnss.lat_ns;
    result->longitudinal    = lla[1];
    result->longitudinal_ew = pgnss.lon_ew;
    result->time            = pgnss.timestamps;
    p_pdr->NoGnssCount++;
}

/**----------------------------------------------------------------------
* Function    : calPdrHeadingOffset
* Description : GPSźŽϺʱƫƼ㷽ƫ
* Date        : 2020/07/08 logzhan
*---------------------------------------------------------------------**/
void CalPdrHeadingOffset(Nmea_t* nmea_data, PDR_t* p_pdr) {
    //ߵ̬GPSнǶȵĶ׼жGPSźǷԼߵǷƽȣ
    if (refGpsYaw != -1000)return;

    if ((pgnss.satellites_num > 4 && pgnss.HDOP < 2.0 && pgnss.yaw != -1 && pgnss.vel >= 1.0)
        && (nmea_data->accuracy.err > 0 && nmea_data->accuracy.err < 4)
        && (DetUserStatic(pdr, &pgnss, (p_pdr->steps - p_pdr->lastSteps)) == 0 && pdr->MotionType == 2))
    {
        /*̬GPS׼part1 start*/
        double imuHeading = atan2(2 * (Ahrs.qnb[0] * Ahrs.qnb[3] + Ahrs.qnb[1] * Ahrs.qnb[2]), 
            1 - 2 * (Ahrs.qnb[2] * Ahrs.qnb[2] + Ahrs.qnb[3] * Ahrs.qnb[3]));
        if (imuHeading < 0) {
            imuHeading += TWO_PI;
        }

        if (fabs(imuHeading - p_pdr->GpsHeading) < PI / 18)
        {
            //printf("offset 궨\n");
            //p_pdr->insHeadingOffset = imuHeading - p_pdr->GpsHeading;
        }
    }
}

/**----------------------------------------------------------------------
* Function    : DetectFixMode
* Description : ⵱ǰPDRڵģʽǳغ;ֹģʽ,ѡ
*               ԭʼGPS߲
* Date        : 2020/07/08 logzhan : ޸-1ΪINVAILD_GPS_YAW
*               Ŀɶ
*---------------------------------------------------------------------**/
int DetectFixMode(GNSS_t* pgnss, EKFPara_t* kf, PDR_t* pdr, LctFs_t* result) {
    // ûǷھֹ״̬
    if (DetUserStatic(pdr, pgnss, pdr->deltaStepsPerGps)) {
        if (pgnss->error > 0 && pgnss->error < 15 &&
            pgnss->lastError >= pgnss->error) {
            OutputRawGnssPos(pgnss, result);
            pgnss->lastError = pgnss->error;
            return TYPE_FIX_PDR;
        }
        // λ
        return TYPE_FIX_NONE;
    }
    // Ǽ⵽ǿ״̬GPSλ
    if (DetectCarMode(pgnss, pdr, pdr->deltaStepsPerGps, &pgnss->overVelCount)) {
        // ⵽ģʽPDR״̬Ҫ³ʼ
        ResetSystemStatus(kf);
        if (pgnss->error >= 0 || (pgnss->satellites_num > 4
            && pgnss->HDOP < 2.0)) {
            OutputRawGnssPos(pgnss, result);
            return TYPE_FIX_GPS;
        }
        // ȹλ
        return TYPE_FIX_NONE;
    }
    return TYPE_FIX_PDR;
}


/**----------------------------------------------------------------------
* Function    : GnssCalHeading
* Description : ȡGPSƫ(0 - 360)
* Date        : 2022/09/16 logzhan
*               Ŀɶ
*---------------------------------------------------------------------**/
double GnssCalHeading(GNSS_t* pgnss) {
    double gnssHeading = INVAILD_GPS_YAW;
    if (pgnss->satellites_num > 4 && pgnss->HDOP < 2.0 && pgnss->vel >= 1.0
        && pgnss->yaw != INVAILD_GPS_YAW && pgnss->error < 50) {
        gnssHeading = pgnss->yaw * RADIAN_PER_DEG;
    }
    return gnssHeading;
}

/**----------------------------------------------------------------------
* Function    : pdr_nmea2Gnss
* Description : nmeaݽṹתgnss
* Date        : 2020/07/08 logzhan
*---------------------------------------------------------------------**/
void Nmea2Gnss(Nmea_t* nmea_data, GNSS_t* pgnss) 
{
    NmeaRMC_t rmc = { 0 };
    if (nmea_data->rmc[0].time > nmea_data->rmc[1].time) {
        rmc = nmea_data->rmc[0];
    }else {
        rmc = nmea_data->rmc[1];
    }
    // һεAccuracy Error
    pgnss->lastError      = pgnss->error;
    pgnss->lat            = rmc.latitude * PI / 180;
    pgnss->lat_ns         = rmc.rmc_latitude_ns;
    pgnss->lon            = rmc.longitudinal * PI / 180;
    pgnss->lon_ew         = rmc.longitudinal_ew;
    //pgnss->vel            = rmc.speed * GPS_SPEED_UNIT;
    pgnss->vel            = rmc.speed;
    pgnss->HDOP           = nmea_data->gsa[0].hdop;
    pgnss->yaw            = rmc.cog;
    pgnss->quality        = 0;
    pgnss->error          = nmea_data->accuracy.err;
    pgnss->timestamps     = rmc.time;
    pgnss->satellites_num = nmea_data->gga.satellites_nb;

}
/**----------------------------------------------------------------------
* Function    : OutputRawGnssPos
* Description : GPSλ
* Date        : 2022/10/15 logzhan
*---------------------------------------------------------------------**/
void OutputRawGnssPos(GNSS_t* pgnss, LctFs_t* result) {
    result->latitude = pgnss->lat;
    result->latitude_ns = pgnss->lat_ns;
    result->longitudinal = pgnss->lon;
    result->longitudinal_ew = pgnss->lon_ew;
    result->time = pgnss->timestamps;
}


/**----------------------------------------------------------------------
* Function    : ResetOutputLoction
* Description : 1ʼʼλá˲ʼ
*               2GPSֵֵresult
* Date        : 2022/09/19 logzhan
*---------------------------------------------------------------------**/
int ResetOutputLoction(GNSS_t* pgnss, PDR_t* g_pdr, EKFPara_t* kf, LctFs_t* result) {

    double stepLength = 0.7;
    double ned[3] = { 0 };

    g_pdr->PllaInit[0] = pgnss->lat;
    g_pdr->PllaInit[1] = pgnss->lon;
    g_pdr->PllaInit[2] = 0;
    WGS842NED(g_pdr->PllaInit, g_pdr->PllaInit, ned);

    g_pdr->x0 = ned[0];
    g_pdr->y0 = ned[1];

    kf->pXk[0] = ned[0];
    kf->pXk[1] = ned[1];
    kf->pXk[2] = stepLength;
    // GPSʱGPSĺǲһ׼
    kf->pXk[3] = pgnss->yaw * RADIAN_PER_DEG;

    for (uchar i = 0; i < N; i++) {
        kf->Xk[i] = kf->pXk[i];
    }
    // QR
    for (int i = 0; i < N; i++) {
        for (int l = 0; l < N; l++) {
            kf->Q[i][l] = 0.0;
            kf->R[i][l] = 0.0;
        }
    }
    // GPSλý
    OutputRawGnssPos(pgnss, result);
    g_pdr->NoGnssCount = 0;
    return PDR_TRUE;
}

/**----------------------------------------------------------------------
* Function    : EkfGnssInsLocFusion
* Description : EKFGNSSINSں϶λ
* Date        : 2022/09/21 logzhan
*---------------------------------------------------------------------**/
void EkfGnssInsLocFusion(GNSS_t* pgnss, PDR_t* pdr, Classifer_t* sys, double yaw_bias,
    EKFPara_t* kf, LctFs_t* res) {
    double plla[3] = { 0 };
    double Ned[3] = { 0 };
    double yaw_thr = 6;

    int scene_type = pdr->sceneType;

    plla[0] = pgnss->lat;
    plla[1] = pgnss->lon;
    plla[2] = 0;

    WGS842NED(plla, pdr->PllaInit, Ned);

    pgnss->xNed = Ned[0];
    pgnss->yNed = Ned[1];

    //ںϲ
    EKFStateUpdate(kf, pgnss, sys, pdr, scene_type);

    for (uchar i = 0; i < N; i++) {
        for (uchar l = 0; l < N; l++) {
            kf->pPk[i][l] = kf->Pk[i][l];
        }
        kf->pXk[i] = kf->Xk[i];
    }

    Ned[0] = kf->pXk[0] - pdr->x0;
    Ned[1] = kf->pXk[1] - pdr->y0;
    Ned[2] = 0;

    NED2WGS84(pdr->PllaInit, Ned, plla);

    res->latitude = plla[0];
    res->latitude_ns = pgnss->lat_ns;
    res->longitudinal = plla[1];
    res->longitudinal_ew = pgnss->lon_ew;
    res->time = pgnss->timestamps;

    pdr->NoGnssCount = 0;
}

/**----------------------------------------------------------------------
* Function    : PDRInitialAlignment
* Description : PDRʼ׼
* Date        : 2022/09/19 logzhan
*---------------------------------------------------------------------**/
int GnssInsInitialAlignment(GNSS_t* pgnss, EKFPara_t* kf, PDR_t* g_pdr, LctFs_t* result)
{
    double stepLen = 0.7;      // ʼ˶Ϊ0.7m
    if (pgnss->satellites_num > 4 && pgnss->HDOP < 2.0 && pgnss->vel >= 1.5 && pgnss->yaw != -1) {
        // plla λԭ(λǾγ)

        g_pdr->PllaInit[0] = pgnss->lat;
        g_pdr->PllaInit[1] = pgnss->lon;
        g_pdr->PllaInit[2] = 0;

        // plla ת ned
        double ned[3] = { 0 };
        WGS842NED(g_pdr->PllaInit, g_pdr->PllaInit, ned);

        // NEDϵļ, λ
        g_pdr->x0 = ned[0];
        g_pdr->y0 = ned[1];

        // ʼ˲Ĺ۲
        kf->pXk[0] = ned[0];                        // ״̬1: x
        kf->pXk[1] = ned[1];                        // ״̬2y
        kf->pXk[2] = stepLen;                       // ״̬3
        kf->pXk[3] = pgnss->yaw * RADIAN_PER_DEG;   // ״̬4

        for (uchar i = 0; i < N; i++) {
            kf->Xk[i] = kf->pXk[i];
        }

        kf->initHeading = kf->pXk[3];
        for (int i = 0; i < N; i++) {
            for (int l = 0; l < N; l++) {
                kf->Q[i][l] = 0.0;
                kf->R[i][l] = 0.0;
            }
        }
        OutputRawGnssPos(pgnss, result);
        g_pdr->sysStatus = IS_NORMAL;

        return PDR_TRUE;
    }

    if (pgnss->satellites_num > 4 && pgnss->HDOP < 2.0 && pgnss->vel > 0 && pgnss->yaw != -1) {
        OutputRawGnssPos(pgnss, result);
        return PDR_TRUE;
    }
    return PDR_FALSE;
}